local addonName, addon = ...

local ldb = LibStub:GetLibrary("LibDataBroker-1.1")
local module = {name = addonName .. "-CombatStats"}

local errorHandler = geterrorhandler()

function module:Enable()
    -- Don't load or register anything if this plugin is disabled
    if addon:IsPluginDisabled(self.name) then
        return
    end

    self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", "EventHandler")
    self:RegisterEvent("UNIT_POWER_FREQUENT", "EventHandler")
    self:RegisterEvent("UNIT_SPELLCAST_SENT", "EventHandler")
    self:RegisterEvent("PLAYER_MOUNT_DISPLAY_CHANGED", "EventHandler")
    self:RegisterEvent("UNIT_AURA", "EventHandler")

    self.dirty = true
    self.timerScheduled = false
    self.timerDelay = 0.25

    self.timerCallback = function()
        self:TimerCallback()
    end

    self.percFormat = "%.2f%%"
    
    self.colorPalette = {
        "FFC5FFFD",
        "FF88D9E6",
        "FF8B8BAE",
        "FFC7D1CC",
        "FFA1BAB8",
    }

    local secondaries = {
        {label = "Crit", token = "{critp}", value = self.tokenMapping.critp()},
        {label = "Haste", token = "{hastep}", value = self.tokenMapping.hastep()},
        {label = "Mastery", token = "{masteryp}", value = self.tokenMapping.masteryp()},
        {label = "Vers", token = "{versoffp}", value = self.tokenMapping.versoffp()},
    }

    table.sort(secondaries, function(a,b)
        return a.value >= b.value
    end)

    self.textFormat = self:GetFormatString(string.format("|c%s%s|r: %s, |c%s%s|r: %s",
        self.colorPalette[1],
        secondaries[1].label,
        secondaries[1].token,
        self.colorPalette[2],
        secondaries[2].label,
        secondaries[2].token))

    self:CreateDataFeed()
end

function module:CreateDataFeed()
    self.dataFeed = ldb:NewDataObject(self.name, {
        type = "data source",
        icon = "Interface\\Icons\\icon_upgradestone_humanoid_rare",
        text = "Speed: X",
        OnTooltipShow = function(tooltip)
            self:OnTooltipShow(tooltip)
        end,
    })
end

module.tokenMapping = {
    primary = function() return module:CalcPrimaryStat() end,
    primaryname = function()
        local value, name = module:CalcPrimaryStat()
        return name
    end,
    speed = function() return module:CalcSpeed() end,
    haste = function() return GetCombatRating(CR_HASTE_MELEE) end,
    hastep = function() return GetHaste() end,
    dodge = function() return GetDodgeChance() end,
    masteryp = function() return GetMasteryEffect() end,
    mastery = function() return GetCombatRating(CR_MASTERY) end,
    critp = function() return GetSpellCritChance(2) end,
    crit = function() return GetCombatRating(CR_CRIT_MELEE) end,
    leechp = function() return GetLifesteal() end,
    leech = function() return GetCombatRating(CR_LIFESTEAL) end,
    block = function() return GetBlockChance() end,
    armor = function() return select(2, UnitArmor("player")) end,
    avoidancep = function() return GetCombatRatingBonus(CR_AVOIDANCE) end,
    avoidance = function() return GetCombatRating(CR_AVOIDANCE) end,
    parryp = function() return GetParryChance() end,
    parry = function() return GetCombatRating(CR_PARRY) end,
    vers = function() return GetCombatRating(CR_VERSATILITY_DAMAGE_DONE) end,
    versoffp = function()
        return GetCombatRatingBonus(CR_VERSATILITY_DAMAGE_DONE)
    end,
    versdefp = function() return GetCombatRatingBonus(CR_VERSATILITY_DAMAGE_TAKEN) end,
}
local tokenMapping = module.tokenMapping

function module.GetTokenValue(token)
    token = token:lower()
    if tokenMapping[token] then
        local ok, value = xpcall(tokenMapping[token], errorHandler)
        if type(value) == "string" then
            return value
        elseif type(value) == "number" then
            local f, e = math.modf(value)
            if e == 0 then
                return string.format("%d", value)
            else
                return string.format(module.percFormat, value)
            end            
        end
    end

    return "??"
end

function module:GetFormatString(s)
    local formatted = string.gsub(s, "{([^{}]*)}", self.GetTokenValue)
    return formatted
end

function module:TimerCallback()
    self.timerScheduled = false
    self.dataFeed.text = self:GetFormatString(self.textFormat)
end

function module:CalcSpeed()
    local result
    local cur, run, fly, swim = GetUnitSpeed("player");
    if cur ~= 0 then
        result = cur
    elseif IsFlying() then
        cur = fly
    elseif IsSwimming() then
        cur = swim
    else
        cur = run
    end

    if cur then
        return cur / 7 * 100
    else
        return 0
    end
end

function module:CalcPrimaryStat()
    local currentSpec = GetSpecialization()
    local primaryToken
    local effectiveStat

    if currentSpec then
        local id, _, _, _, _, mainStatID = GetSpecializationInfo(currentSpec)
        local roleToken = id and GetSpecializationRoleByID(id) or nil

        if mainStatID == 1 then
            primaryToken = "Strength"
        elseif mainStatID == 2 then
            primaryToken = "Agility"
        elseif mainStatID == 4 then
            primaryToken = "Intellect"
        end

        effectiveStat = select(2, UnitStat("player", mainStatID))
    end

    if not effectiveStat then
        effectiveStat = 0
    end

    return effectiveStat, primaryToken
end

function module:EventHandler(event, unit, subEvent, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, spellId, ...)
    local dirty = false

    if event == "COMBAT_LOG_EVENT_UNFILTERED" and  destGUID ~= UnitGUID("player") then
        dirty = true
    elseif event == "UNIT_POWER_FREQUENT" then
        dirty = true
    elseif event == "UNIT_SPELLCAST_SENT" and ((unit and unit ~= "player")) then
        dirty = true
    elseif event == "PLAYER_MOUNT_DISPLAY_CHANGED" or (event == "UNIT_AURA" and unit == "player") then
        dirty = true
    end

    module.dirty = dirty
    if module.dirty and not module.timerScheduled then
        module.timerScheduled = true
        C_Timer.After(self.timerDelay, self.timerCallback)
    end
end    

function module:OnTooltipShow(tooltip)
    if not tooltip or not tooltip.AddLine then return end

    local colorIdx = 1
    local getCurrentColor = function()
        return self.colorPalette[colorIdx]
    end
    local getNextColor = function()
        local result = self.colorPalette[colorIdx]
        colorIdx = colorIdx + 1
        if colorIdx > #self.colorPalette then
            colorIdx = 1
        end
        return result
    end

    tooltip:AddLine(module.name)
    tooltip:AddDoubleLine(
        self:GetFormatString(string.format("|c%s{primaryname}|r:", getNextColor())),
        self:GetFormatString("|cffffffff{primary}|r"))
    tooltip:AddDoubleLine(
        self:GetFormatString(string.format("|c%sSpeed|r:", getNextColor())),
        self:GetFormatString(string.format("|c%s{speed}|r", getCurrentColor())))
    tooltip:AddDoubleLine(
        self:GetFormatString(string.format("|c%sHaste (|cffffffff{haste}|r)|r:", getNextColor())),
        self:GetFormatString(string.format("|c%s{hastep}|r", getCurrentColor())))
    tooltip:AddDoubleLine(
        self:GetFormatString(string.format("|c%sCrit (|cffffffff{crit}|r)|r:", getNextColor())),
        self:GetFormatString(string.format("|c%s{critp}|r", getCurrentColor())))
    tooltip:AddDoubleLine(
        self:GetFormatString(string.format("|c%sMastery (|cffffffff{mastery}|r)|r:", getNextColor())),
        self:GetFormatString(string.format("|c%s{masteryp}|r", getCurrentColor())))
    tooltip:AddDoubleLine(
        self:GetFormatString(string.format("|c%sVers (|cffffffff{vers}|r)|r:", getNextColor())),
        self:GetFormatString(string.format("|c%s{versoffp} / {versdefp}%%|r", getCurrentColor())))
    tooltip:AddDoubleLine(
        self:GetFormatString(string.format("|c%sLeech (|cffffffff{leech}|r)|r:", getNextColor())),
        self:GetFormatString(string.format("|c%s{leechp}|r", getCurrentColor())))
    tooltip:AddDoubleLine(
        self:GetFormatString(string.format("|c%sAvoidance (|cffffffff{avoidance}|r)|r:", getNextColor())),
        self:GetFormatString(string.format("|c%s{avoidancep}|r", getCurrentColor())))
    tooltip:AddDoubleLine(
        self:GetFormatString(string.format("|c%sParry (|cffffffff{parry}|r)|r:", getNextColor())),
        self:GetFormatString(string.format("|c%s{parryp}|r", getCurrentColor())))
end

addon:RegisterModule(module, module.name)